@@ -45,7 +45,7 @@ PERFORMANCE OF THIS SOFTWARE.
  *
  */
 
-static char __version__[] = "1.1";
+static char __version__[] = "1.2";
 
 static char pymqe_doc[] = " \
 pymqe - A Python MQ Extension.  This presents a low-level Python \
@@ -124,6 +124,11 @@ static PyObject *ErrorObj;
 #ifdef PYMQI_FEATURE_SSL
 #define PYMQI_MQSCO_SIZEOF sizeof(MQSCO)
 #endif
+#ifdef MQCMDL_LEVEL_700
+#define PYMQI_MQSD_SIZEOF sizeof(MQSD)
+#define PYMQI_MQSRO_SIZEOF sizeof(MQSRO)
+#endif
+
 
 /* ----------------------------------------------------- */
 
@@ -695,7 +700,7 @@ static PyObject *pymqe_MQSET(PyObject *self, PyObject *args) {
       return NULL;
     }
     strncpy(charAttrs, PyString_AsString(attrArg), MAX_CHARATTR_LENGTH);
-    charAttrCount = (MQLONG)strlen(charAttrs); // xx
+    charAttrCount = (MQLONG)strlen(charAttrs);
   }
 
   Py_BEGIN_ALLOW_THREADS
@@ -705,6 +710,51 @@ static PyObject *pymqe_MQSET(PyObject *self, PyObject *args) {
   return Py_BuildValue("(ll)", (long) compCode, (long) compReason);
 }
 
+/* Publish/subscribe - Hannes Wagener 2011 */
+#ifdef MQCMDL_LEVEL_700
+
+static char pymqe_MQSUB__doc__[] =
+"MQSUB(connectionHandle, sd, objectHandle, subHandle, compCode, reasonCode) \
+ \
+Calls the MQI MQSUB(connectionHandle, subDesc, objectHandle, subHandle, compCode, reasonCode) \
+";
+
+static PyObject * pymqe_MQSUB(PyObject *self, PyObject *args) {
+  MQSD *subDescP;
+  MQHOBJ subHandle;
+  MQHOBJ objectHandle;
+  MQLONG compCode, compReason;
+  PyObject *rv;
+
+  char *subDescBuffer;
+  int subDescBufferLength;
+
+
+  long lQmgrHandle;
+
+  if (!PyArg_ParseTuple(args, "ls#l", &lQmgrHandle,
+            &subDescBuffer, &subDescBufferLength,
+            &objectHandle)) {
+    return NULL;
+  }
+  if (checkArgSize(subDescBufferLength, PYMQI_MQSD_SIZEOF, "MQSD")) {
+    return NULL;
+  }
+
+  subDescP = (MQSD *)subDescBuffer;
+
+  Py_BEGIN_ALLOW_THREADS
+  MQSUB((MQHCONN) lQmgrHandle, subDescP, &objectHandle, &subHandle,
+    &compCode, &compReason);
+  Py_END_ALLOW_THREADS
+
+  rv = Py_BuildValue("(s#llll)", subDescP, PYMQI_MQSD_SIZEOF, objectHandle, subHandle,
+             (long) compCode, (long) compReason);
+  return rv;
+
+}
+
+#endif /* MQCMDL_LEVEL_700 */
 
 #ifdef PYMQI_FEATURE_MQAI
 
@@ -719,6 +769,7 @@ static void cleanupBags(MQHBAG adminBag, MQHBAG responseBag) {
   }
 }
 
+
 static char pymqe_mqaiExecute__doc__[] =
 "mqaiExecute(qMgr, cmd, args)\
 \
@@ -741,12 +792,14 @@ static PyObject *pymqe_mqaiExecute(PyObject *self, PyObject *args) {
     MQHBAG responseBag = MQHB_UNUSABLE_HBAG;
     MQHBAG resultBag;
     PyObject *resultsList;
-
     PyObject *pyItemIntVal;
     PyObject *pyItemStrVal;
     PyObject *returnValue;
+    PyObject *resultDictValue;
+    PyObject *newList;
 
     long lQmgrHandle, lCmdCode;
+    int isByteString = 0;
 
     if(!PyArg_ParseTuple(args, "ll|O", &lQmgrHandle, &lCmdCode, &argDict)) {
         return NULL;
@@ -789,8 +842,9 @@ static PyObject *pymqe_mqaiExecute(PyObject *self, PyObject *args) {
               MQLONG paramType = 0;
               MQLONG intArg = 0;
               MQCHAR *strArg = "";
+              MQBYTE *strByteArg = NULL;
 
-              int nKeys = (int)PyList_Size(argKeys); /* Owned ref */ // xx
+              int nKeys = (int)PyList_Size(argKeys); /* Owned ref */
               int i;
 
               for(i = 0; i < nKeys; i++) {
@@ -826,15 +880,30 @@ static PyObject *pymqe_mqaiExecute(PyObject *self, PyObject *args) {
                        strArg = PyString_AsString(value);
                        mqAddString(adminBag, paramType, MQBL_NULL_TERMINATED, strArg, &compCode, &compReason);
                   } else {
-                      PyObject *keyStr = PyObject_Str(key);    /* Owned ref */
-                      PyObject *valStr = PyObject_Str(value);  /* Owned ref */
-                      PyErr_Format(ErrorObj, "Value %s for key %s is neither a long or string",
-                                   PyString_AsString(valStr), PyString_AsString(keyStr));
-                      Py_XDECREF(keyStr);
-                      Py_XDECREF(valStr);
-                      Py_DECREF(resultsList);
-                      cleanupBags(adminBag, responseBag);
-                      return NULL;
+                      isByteString = PyObject_HasAttrString(value, "pymqi_byte_string");
+                      if(1 == isByteString) {
+                        /* value is a ByteString.  have to use its "value" attribute */
+                        PyObject *byteStringValue;
+                        byteStringValue = PyObject_GetAttrString(value, "value"); /* Owned ref */
+                        strByteArg = (MQBYTE *)PyString_AsString(byteStringValue);
+
+#ifdef MQCMDL_LEVEL_700
+                        mqAddByteString(adminBag, paramType, PyObject_Length(value), strByteArg, &compCode, &compReason);
+#endif /* MQCMDL_LEVEL_700 */
+
+                        Py_XDECREF(byteStringValue);
+                        }
+                      else {
+                        PyObject *keyStr = PyObject_Str(key);    /* Owned ref */
+                        PyObject *valStr = PyObject_Str(value);  /* Owned ref */
+                        PyErr_Format(ErrorObj, "Value %s for key %s is not a long, string nor a pymqi.ByteString instance",
+                                     PyString_AsString(valStr), PyString_AsString(keyStr));
+                        Py_XDECREF(keyStr);
+                        Py_XDECREF(valStr);
+                        Py_DECREF(resultsList);
+                        cleanupBags(adminBag, responseBag);
+                        return NULL;
+                      }
                  }
                  if(compCode != MQCC_OK) {
                      break;
@@ -953,7 +1022,28 @@ static PyObject *pymqe_mqaiExecute(PyObject *self, PyObject *args) {
                                         mqInquireInteger(attrsBag, MQSEL_ANY_USER_SELECTOR, j, &itemIntVal, &compCode, &compReason);
 
                                         pyItemIntVal = PyLong_FromLong(itemIntVal);
-                                        PyDict_SetItem(resultsDict, key, pyItemIntVal);
+
+                                        if (PyDict_Contains(resultsDict, key) > 0)
+                                        {
+                                             resultDictValue = PyDict_GetItem(resultsDict, key);
+                                             if (PyList_Check(resultDictValue))
+                                              {
+                                                 PyList_Append(resultDictValue,pyItemIntVal);
+                                                 PyDict_SetItem(resultsDict, key, resultDictValue);
+                                             }
+                                             else
+                                             {
+                                                 newList = PyList_New(0);
+                                                 PyList_Append(newList, resultDictValue);
+                                                 PyList_Append(newList, pyItemIntVal);
+                                                 PyDict_SetItem(resultsDict, key, newList);
+                                                 Py_XDECREF(newList);
+                                                }
+                                        }
+                                        else
+                                        {
+                                            PyDict_SetItem(resultsDict, key, pyItemIntVal);
+                                        }
 
                                         Py_XDECREF(pyItemIntVal);
                                         Py_XDECREF(key);
@@ -997,18 +1087,83 @@ static PyObject *pymqe_mqaiExecute(PyObject *self, PyObject *args) {
                                         itemStrVal[strLength] = 0;
 
                                         pyItemStrVal = PyString_FromString(itemStrVal);
-                                        PyDict_SetItem(resultsDict, key, pyItemStrVal);
+
+                                        if (PyDict_Contains(resultsDict, key) > 0) {
+                                              resultDictValue = PyDict_GetItem(resultsDict, key);
+                                              if (PyList_Check(resultDictValue)) {
+                                                  PyList_Append(resultDictValue,pyItemStrVal);
+                                                  PyDict_SetItem(resultsDict, key, resultDictValue);
+                                              }
+                                              else {
+                                                  newList = PyList_New(0);
+                                                  PyList_Append(newList, resultDictValue);
+                                                  PyList_Append(newList, pyItemStrVal);
+                                                  PyDict_SetItem(resultsDict, key, newList);
+                                                  Py_XDECREF(newList);
+                                              }
+                                        }
+                                        else {
+                                          PyDict_SetItem(resultsDict, key, pyItemStrVal);
+                                        }
 
                                         free(itemStrVal);
                                         Py_XDECREF(key);
                                         Py_XDECREF(pyItemStrVal);
 
                                     }
+
+#ifdef MQCMDL_LEVEL_700
+                                    else if(itemType == MQITEM_BYTE_STRING) {
+
+                                        MQBYTE *itemByteStrVal;
+                                        MQLONG byteStrLength;
+                                        /*
+                                         * Two calls are needed - one to get the byte string
+                                         * length, and one to get the byte string itself.
+                                         */
+
+                                        /* 1st call */
+                                        mqInquireByteString(attrsBag, MQSEL_ANY_USER_SELECTOR, j, 0, 0, &byteStrLength, &compCode, &compReason);
+
+                                        if(compCode != MQCC_OK && compReason != MQRC_STRING_TRUNCATED) {
+                                            break;
+                                        }
+
+                                        byteStrLength++;   /* + one for the Null */
+
+                                        if(!(itemByteStrVal = malloc(byteStrLength))) {
+                                            PyErr_SetString(ErrorObj, "Out of memory");
+                                            cleanupBags(adminBag, responseBag);
+                                            Py_DECREF(resultsList);
+                                            Py_XDECREF(resultsDict);
+                                            Py_XDECREF(key);
+                                            return NULL;
+                                        }
+
+                                        /* 2nd call */
+                                        mqInquireByteString(attrsBag, MQSEL_ANY_USER_SELECTOR, j, byteStrLength, itemByteStrVal,
+                                                        &byteStrLength, &compCode, &compReason);
+
+                                        if(compCode != MQCC_OK) {
+                                            break;
+                                        }
+
+                                        /*byte strings may contain nulls */
+                                        pyItemStrVal = PyString_FromStringAndSize((char *)itemByteStrVal, byteStrLength);
+                                        PyDict_SetItem(resultsDict, key, pyItemStrVal);
+
+                                        free(itemByteStrVal);
+                                        Py_XDECREF(key);
+                                        Py_XDECREF(pyItemStrVal);
+
+                                    }
+#endif /* MQCMDL_LEVEL_700 */
+
                                     else {
                                         /*
                                          * Must be a bag. What to do? Maybe recurse into it?
                                          */
-                                        PyErr_SetString(ErrorObj, "Bag in a Bag. Send clue to les@open-networks.co.uk");
+                                        PyErr_SetString(ErrorObj, "Bag in a Bag. Send clue to http://packages.python.org/pymqi/support-consulting-contact.html");
                                         cleanupBags(adminBag, responseBag);
                                         Py_XDECREF(key);
                                         Py_DECREF(resultsList);
@@ -1037,9 +1192,6 @@ static PyObject *pymqe_mqaiExecute(PyObject *self, PyObject *args) {
 
     return returnValue;
 }
-
-
-
 #endif
 
 
@@ -1061,6 +1213,9 @@ static struct PyMethodDef pymqe_methods[] = {
   {"MQSET", (PyCFunction)pymqe_MQSET, METH_VARARGS, pymqe_MQSET__doc__},
 #ifdef  PYMQI_FEATURE_MQAI
   {"mqaiExecute", (PyCFunction)pymqe_mqaiExecute, METH_VARARGS, pymqe_mqaiExecute__doc__},
+#endif
+#ifdef MQCMDL_LEVEL_700
+  {"MQSUB", (PyCFunction)pymqe_MQSUB, METH_VARARGS, pymqe_MQSUB__doc__},
 #endif
   {NULL, (PyCFunction)NULL, 0, NULL}        /* sentinel */
 };
@@ -1132,4 +1287,3 @@ __declspec(dllexport)
   if (PyErr_Occurred())
     Py_FatalError("can't initialize module pymqe");
 }
-
