@@ -11,7 +11,6 @@
 # provided "AS-IS" without warranty of any kind, either express or
 # implied. So there.
 #
-
 """
 PyMQI - Python MQI Wrapper Classes
 
@@ -25,17 +24,22 @@ structure members by attribute, dictionary etc.
 The classes are:
 
     * MQOpts - Base class for defining MQI parameter structures.
-    * gmo - MQI MQGMO structure class.
-    * pmo - MQI MQPMO structure class.
-    * od - MQI MQOD structure class.
-    * md - MQI MQMD structure class.
-    * cd - MQI MQCD structure class
-    * sco - MQI MQSCO structure class
-
-    * QueueManager - Queue Manager operations.
+    * CD - MQI MQCD structure class
+    * MD - MQI MQMD structure class
+    * GMO - MQI MQGMO structure class.
+    * OD - MQI MQOD structure class.
+    * PMO - MQI MQPMO structure class.
+    * RFH2 - MQI MQRFH2 structure class.
+    * SCO - MQI MQSCO structure class
+    * SRO - MQI MQSRO structure class
+    * SD - MQI MQSD structure class
+    * TM - MQI MQTM structure class
+    * TMC2- MQI MQTMC2 structure class
+    * QueueManager - Queue Manager operations
     * Queue - Queue operations
+    * Topic - Publish/subscribe topic operations
+    * Subscription - Publish/subscribe subscription operations
     * PCFExecute - Programmable Command Format operations
-
     * Error - Base class for pymqi errors.
     * MQMIError - MQI specific error
     * PYIFError - Pymqi error
@@ -44,12 +48,13 @@ The following MQI operations are supported:
 
     * MQCONN, MQDISC (QueueManager.connect()/QueueManager.disconnect())
     * MQCONNX (QueueManager.connectWithOptions())
-    * MQOPEN/MQCLOSE (Queue.open(), Queue.close())
+    * MQOPEN/MQCLOSE (Queue.open(), Queue.close(), Topic.open(), Topic.close())
     * MQPUT/MQPUT1/MQGET (Queue.put(), QueueManager.put1(), Queue.get())
     * MQCMIT/MQBACK (QueueManager.commit()/QueueManager.backout())
     * MQBEGIN (QueueuManager.begin())
     * MQINQ (QueueManager.inquire(), Queue.inquire())
     * MQSET (Queue.set())
+    * MQSUB (Subscription.sub())
     * And various MQAI PCF commands.
 
 The supported command levels (from 5.0 onwards) for the version of MQI
@@ -82,12 +87,23 @@ their MQI counterparts.
 """
 
 # Stdlib
-import struct, exceptions, types, threading
+import struct
+import exceptions
+import types
+import threading
+import ctypes
+# import xml parser.  lxml/etree only available since python 2.5
+use_minidom = False
+try:
+    import lxml.etree
+except Exception:
+    from xml.dom.minidom import parseString
+    use_minidom = True
 
 # PyMQI
 import pymqe, CMQC, CMQCFC, CMQXC
 
-__version__ = "1.1"
+__version__ = "1.2"
 __mqlevels__ = pymqe.__mqlevels__
 __mqbuild__ = pymqe.__mqbuild__
 
@@ -273,6 +289,80 @@ class MQOpts:
         Return the packed buffer as a printable string."""
         return str(self.pack())
 
+    def get_length(self):
+        """get_length()
+
+        Returns the length of the (would be) packed buffer.
+
+        """
+
+        return struct.calcsize(self.__format)
+
+    def set_vs(self, vs_name, vs_value=None, vs_offset=0, vs_buffer_size=0,
+               vs_ccsid=0):
+        """set_vs(vs_name, vs_value, vs_offset, vs_buffer_size, vs_ccsid)
+
+        This method aids in the setting of the MQCHARV (variable length
+        string) types in MQ structures.  The type contains a pointer to a
+        variable length string.  A common example of a MQCHARV type
+        is the ObjectString in the MQOD structure.
+        In pymqi the ObjectString is defined as 5 separate
+        elements (as per MQCHARV):
+        ObjectStringVSPtr - Pointer
+        ObjectStringVSOffset - Long
+        ObjectStringVSBufSize - Long
+        ObjectStringVSLength - Long
+        ObjectStringVSCCSID - Long
+
+        """
+
+        vs_name_vsptr = ""
+        #if the VSPtr name is passed - remove VSPtr to be left with name.
+        if vs_name.endswith("VSPtr"):
+            vs_name_vsptr = vs_name
+        else:
+            vs_name_vsptr = vs_name + "VSPtr"
+
+        vs_name_vsoffset = vs_name + "VSOffset"
+        vs_name_vsbuffsize = vs_name + "VSBufSize"
+        vs_name_vslength = vs_name + "VSLength"
+        vs_name_vsccsid = vs_name + "VSCCSID"
+
+        c_vs_value = None
+        c_vs_value_p = 0
+
+        if vs_value is not None:
+            c_vs_value = ctypes.create_string_buffer(vs_value)
+            c_vs_value_p = ctypes.cast(c_vs_value, ctypes.c_void_p).value
+
+        self[vs_name_vsptr] = c_vs_value_p
+        self[vs_name_vsoffset] = vs_offset
+        self[vs_name_vsbuffsize] = vs_buffer_size
+        self[vs_name_vslength] = len(vs_value)
+        self[vs_name_vsccsid] = vs_ccsid
+
+    def get_vs(self, vs_name):
+        """get_vs(vs_name)
+
+        This method returns the string to which the VSPtr pointer points to.
+
+        """
+
+        vs_name_vsptr = ""
+        #if the VSPtr name is passed - remove VSPtr to be left with name.
+        if vs_name.endswith("VSPtr"):
+            vs_name_vsptr = vs_name
+        else:
+            vs_name_vsptr = vs_name + "VSPtr"
+
+        c_vs_value = None
+        c_vs_value_p = self[vs_name_vsptr]
+        if c_vs_value_p != 0:
+            c_vs_value = ctypes.cast(c_vs_value_p, ctypes.c_char_p).value
+
+        return c_vs_value
+
+
 #
 # Sub-classes of MQOpts representing real MQI structures.
 #
@@ -307,6 +397,9 @@ class gmo(MQOpts):
 
         apply(MQOpts.__init__, (self, tuple(opts)), kw)
 
+# Backward compatibility
+GMO = gmo
+
 
 class pmo(MQOpts):
     """pmo(**kw)
@@ -343,6 +436,9 @@ class pmo(MQOpts):
 
         apply(MQOpts.__init__, (self, tuple(opts)), kw)
 
+# Backward compatibility
+PMO = pmo
+
 class od(MQOpts):
     """od(**kw)
 
@@ -403,6 +499,8 @@ class od(MQOpts):
 
         apply(MQOpts.__init__, (self, tuple(opts)), kw)
 
+# Backward compatibility
+OD = od
 
 class md(MQOpts):
     """md(**kw)
@@ -443,6 +541,245 @@ class md(MQOpts):
             ['MsgFlags', CMQC.MQMF_NONE, MQLONG_TYPE],
             ['OriginalLength', CMQC.MQOL_UNDEFINED, MQLONG_TYPE])), kw)
 
+# Backward compatibility
+MD = md
+
+# RFH2 Header parsing/creation Support - Hannes Wagener - 2010.
+class RFH2(MQOpts):
+    """RFH2(**kw)
+
+    Construct a RFH2 Structure with default values as per MQI. The
+    default values may be overridden by the optional keyword arguments
+    'kw'.  Attempt to parse the RFH2 when unpack is called.
+
+    """
+
+    initial_opts = [['StrucId', CMQC.MQRFH_STRUC_ID, '4s'],
+            ['Version', CMQC.MQRFH_VERSION_2, MQLONG_TYPE],
+            ['StrucLength', 0, MQLONG_TYPE],
+            ['Encoding', CMQC.MQENC_NATIVE, MQLONG_TYPE],
+            ['CodedCharSetId', CMQC.MQCCSI_Q_MGR, MQLONG_TYPE],
+            ['Format', CMQC.MQFMT_NONE, '8s'],
+            ['Flags', 0, MQLONG_TYPE],
+            ['NameValueCCSID', CMQC.MQCCSI_Q_MGR, MQLONG_TYPE]]
+
+    big_endian_encodings = [CMQC.MQENC_INTEGER_NORMAL,
+                            CMQC.MQENC_DECIMAL_NORMAL,
+                            CMQC.MQENC_FLOAT_IEEE_NORMAL,
+                            CMQC.MQENC_FLOAT_S390,
+                            #17
+                            CMQC.MQENC_INTEGER_NORMAL +
+                            CMQC.MQENC_DECIMAL_NORMAL,
+                            #257
+                            CMQC.MQENC_INTEGER_NORMAL +
+                            CMQC.MQENC_FLOAT_IEEE_NORMAL,
+                            #272
+                            CMQC.MQENC_DECIMAL_NORMAL +
+                            CMQC.MQENC_FLOAT_IEEE_NORMAL,
+                            #273
+                            CMQC.MQENC_INTEGER_NORMAL +
+                            CMQC.MQENC_DECIMAL_NORMAL +
+                            CMQC.MQENC_FLOAT_IEEE_NORMAL]
+
+    def __init__(self, **kw):
+        #take a copy of private initial_opts
+        self.opts = [list(x) for x in self.initial_opts]
+        apply(MQOpts.__init__, (self, tuple(self.opts)), kw)
+
+    def add_folder(self, folder_data):
+        """add_folder(folder_data)
+
+        Adds a new XML folder to the RFH2 header.
+        Checks if the XML is well formed and updates self.StrucLength.
+
+        """
+
+        folder_name = None
+        #check that the folder is valid xml and get the root tag name.
+        if use_minidom:
+            try:
+                folder_name = parseString(folder_data). \
+                                            documentElement.tagName
+            except Exception, e:
+                raise PYIFError("RFH2 - XML Folder not well formed. " \
+                                "Exception: %s" % str(e))
+        else:
+            try:
+                folder_name = lxml.etree.fromstring(folder_data).tag
+            except Exception, e:
+                raise PYIFError("RFH2 - XML Folder not well formed. " \
+                                "Exception: %s" % str(e))
+        #make sure folder length divides by 4 - else add spaces
+        folder_length = len(folder_data)
+        remainder = folder_length % 4
+        num_spaces = 0
+        if remainder != 0:
+            num_spaces = 4 - remainder
+            folder_data = folder_data + " " * num_spaces
+            folder_length = len(folder_data)
+
+        self.opts.append([folder_name + "Length", long(folder_length),
+                          MQLONG_TYPE])
+        self.opts.append([folder_name, folder_data, "%is" % folder_length])
+        #save the current values
+        saved_values = self.get()
+        #reinit MQOpts with new fields added
+        apply(MQOpts.__init__, (self, tuple(self.opts)), )
+        #reset the values to the saved values
+        self.set(**saved_values)
+        #calculate the correct StrucLength
+        self["StrucLength"] = self.get_length()
+
+    def pack(self, encoding=None):
+        """pack(encoding)
+
+        Override pack in order to set correct numeric encoding in the format.
+
+        """
+
+        if encoding is not None:
+            if encoding in self.big_endian_encodings:
+                self.opts[0][2] = ">" + self.initial_opts[0][2]
+                saved_values = self.get()
+                #apply the new opts
+                apply(MQOpts.__init__, (self, tuple(self.opts)), )
+                #set from saved values
+                self.set(**saved_values)
+
+        return MQOpts.pack(self)
+
+    def unpack(self, buff, encoding=None):
+        """unpack(buff, encoding)
+
+        Override unpack in order to extract and parse RFH2 folders.
+        Encoding meant to come from the MQMD.
+
+        """
+
+        if buff[0:4] != CMQC.MQRFH_STRUC_ID:
+            raise PYIFError("RFH2 - StrucId not MQRFH_STRUC_ID. Value: %s" %
+                            str(buff[0:4]))
+
+        if len(buff) < 36:
+            raise PYIFError("RFH2 - Buffer too short. Should be 36 bytes or " \
+                            "longer.  Buffer Length: %s" % str(len(buff)))
+        #take a copy of initial_opts and the lists inside
+        self.opts = [list(x) for x in self.initial_opts]
+
+        big_endian = False
+        if encoding is not None:
+            if encoding in self.big_endian_encodings:
+                big_endian = True
+        else:
+            #if small endian first byte of version should be > 'x\00'
+            if buff[4:5] == "\x00":
+                big_endian = True
+        #indicate bigendian in format
+        if big_endian:
+            self.opts[0][2] = ">" + self.opts[0][2]
+        #apply and parse the default header
+        apply(MQOpts.__init__, (self, tuple(self.opts)), )
+        MQOpts.unpack(self, buff[0:36])
+
+        if self['StrucLength'] < 0:
+            raise PYIFError("RFH2 - 'StrucLength' is negative. " \
+                            "Check numeric encoding.")
+
+        if len(buff) > 36:
+            if self['StrucLength'] > len(buff):
+                raise PYIFError("RFH2 - Buffer too short. Expected: " \
+                                "%s Buffer Length: %s" %
+                                (self['StrucLength'], len(buff)))
+
+        #extract only the string containing the xml folders and loop
+        s =  buff[36:self['StrucLength']]
+        while s:
+            #first 4 bytes is the folder length. supposed to divide by 4.
+            len_bytes = s[0:4]
+            folder_length = 0
+            if big_endian:
+                folder_length = struct.unpack(">l", len_bytes)[0]
+            else:
+                folder_length = struct.unpack("l", len_bytes)[0]
+
+            #move on past four byte length
+            s = s[4:]
+            #extract the folder string
+            folder_data = s[:folder_length]
+            #check that the folder is valid xml and get the root tag name
+            folder_name = None
+            #check that the folder is valid xml and get the root tag name.
+            if use_minidom:
+                try:
+                    folder_name = parseString(folder_data). \
+                                                documentElement.tagName
+                except Exception, e:
+                    raise PYIFError("RFH2 - XML Folder not well formed. " \
+                                    "Exception: %s" % str(e))
+            else:
+                try:
+                    folder_name = lxml.etree.fromstring(folder_data).tag
+                except Exception, e:
+                    raise PYIFError("RFH2 - XML Folder not well formed. " \
+                                    "Exception: %s" % str(e))
+            #append folder length and folder string to self.opts types
+            self.opts.append([folder_name + "Length", long(folder_length),
+                              MQLONG_TYPE])
+            self.opts.append([folder_name, folder_data, "%is" %
+                              folder_length])
+            #move on past the folder
+            s = s[folder_length:]
+
+        #save the current values
+        saved_values = self.get()
+        #apply the new opts
+        apply(MQOpts.__init__, (self, tuple(self.opts)), )
+        #set from saved values
+        self.set(**saved_values)
+        #unpack the buffer? - should get same result?
+        #MQOpts.unpack(self, buff[0:self["StrucLength"]])
+
+class TM(MQOpts):
+    """TM(**kw)
+
+    Construct a MQTM Structure with default values as per MQI. The
+    default values may be overridden by the optional keyword arguments 'kw'.
+
+    """
+
+    def __init__(self, **kw):
+        apply(MQOpts.__init__, (self, (
+            ['StrucId', CMQC.MQTM_STRUC_ID, '4s'],
+            ['Version', CMQC.MQTM_VERSION_1, MQLONG_TYPE],
+            ['QName', '', '48s'],
+            ['ProcessName', '', '48s'],
+            ['TriggerData', '', '64s'],
+            ['ApplType', 0, MQLONG_TYPE],
+            ['ApplId', '', '256s'],
+            ['EnvData', '', '128s'],
+            ['UserData', '', '128s'])), kw)
+
+
+class TMC2(MQOpts):
+    """TMC2(**kw)
+
+    Construct a MQTMC2 Structure with default values as per MQI. The
+    default values may be overridden by the optional keyword arguments 'kw'.
+
+    """
+
+    def __init__(self, **kw):
+        apply(MQOpts.__init__, (self, (
+            ['StrucId', CMQC.MQTMC_STRUC_ID, '4s'],
+            ['Version', CMQC.MQTMC_VERSION_2, '4s'],
+            ['QName', '', '48s'],
+            ['ProcessName', '', '48s'],
+            ['TriggerData', '', '64s'],
+            ['ApplType', '', '4s'],
+            ['ApplId', '', '256s'],
+            ['EnvData', '', '128s'],
+            ['UserData', '', '128s'],
+            ['QMgrName', '', '48s'])), kw)
 
 # MQCONNX code courtesy of John OSullivan (mailto:jos@onebox.com)
 # SSL additions courtesy of Brian Vicente (mailto:sailbv@netscape.net)
@@ -583,7 +920,8 @@ class cd(MQOpts):
 
         apply(MQOpts.__init__, (self, tuple(opts)), kw)
 
-
+# Backward compatibility
+CD = cd
 
 # SCO Class for SSL Support courtesy of Brian Vicente (mailto:sailbv@netscape.net)
 class sco(MQOpts):
@@ -610,6 +948,87 @@ class sco(MQOpts):
 
         apply(MQOpts.__init__, (self, tuple(opts)), kw)
 
+# Backward compatibility
+SCO = sco
+
+class SD(MQOpts):
+    """SD(**kw)
+
+    Construct a MQSD Structure with default values as per MQI. The
+    default values may be overridden by the optional keyword arguments
+    'kw'."""
+
+    def __init__(self, **kw):
+        opts = [['StrucId', CMQC.MQSD_STRUC_ID, '4s'],
+                ['Version', CMQC.MQSD_VERSION_1, MQLONG_TYPE],
+                ['Options', CMQC.MQSO_NON_DURABLE, MQLONG_TYPE],
+                ['ObjectName', '', '48s'],
+                ['AlternateUserId', '', '12s'],
+                ['AlternateSecurityId', CMQC.MQSID_NONE, '40s'],
+                ['SubExpiry', CMQC.MQEI_UNLIMITED, MQLONG_TYPE],
+
+                # ObjectString
+                ['ObjectStringVSPtr', 0, 'P'],
+                ['ObjectStringVSOffset', 0L, MQLONG_TYPE],
+                ['ObjectStringVSBufSize', 0L, MQLONG_TYPE],
+                ['ObjectStringVSLength', 0L, MQLONG_TYPE],
+                ['ObjectStringVSCCSID', 0L, MQLONG_TYPE],
+
+                #Subname
+                ['SubNameVSPtr', 0, 'P'],
+                ['SubNameVSOffset', 0L, MQLONG_TYPE],
+                ['SubNameVSBufSize', 0L, MQLONG_TYPE],
+                ['SubNameVSLength', 0L, MQLONG_TYPE],
+                ['SubNameVSCCSID', 0L, MQLONG_TYPE],
+
+                #SubUserData
+                ['SubUserDataVSPtr', 0, 'P'],
+                ['SubUserDataVSOffset', 0L, MQLONG_TYPE],
+                ['SubUserDataVSBufSize', 0L, MQLONG_TYPE],
+                ['SubUserDataVSLength', 0L, MQLONG_TYPE],
+                ['SubUserDataVSCCSID', 0L, MQLONG_TYPE],
+
+                ['SubCorrelId', CMQC.MQCI_NONE, '24s'],
+                ['PubPriority', CMQC.MQPRI_PRIORITY_AS_Q_DEF, MQLONG_TYPE],
+                ['PubAccountingToken', CMQC.MQACT_NONE, '32s'],
+                ['PubApplIdentityData', '', '32s'],
+
+                #SelectionString
+                ['SelectionStringVSPtr', 0, 'P'],
+                ['SelectionStringVSOffset', 0L, MQLONG_TYPE],
+                ['SelectionStringVSBufSize', 0L, MQLONG_TYPE],
+                ['SelectionStringVSLength', 0L, MQLONG_TYPE],
+                ['SelectionStringVSCCSID', 0L, MQLONG_TYPE],
+
+                ['SubLevel', 0, MQLONG_TYPE],
+
+                 #SelectionString
+                ['ResObjectStringVSPtr', 0, 'P'],
+                ['ResObjectStringVSOffset', 0L, MQLONG_TYPE],
+                ['ResObjectStringVSBufSize', 0L, MQLONG_TYPE],
+                ['ResObjectStringVSLength', 0L, MQLONG_TYPE],
+                ['ResObjectStringVSCCSID', 0L, MQLONG_TYPE]]
+
+        apply(MQOpts.__init__, (self, tuple(opts)), kw)
+
+
+class SRO(MQOpts):
+    """SRO(**kw)
+
+    Construct a MQSRO Structure with default values as per MQI. The
+    default values may be overridden by the optional keyword arguments
+    'kw'."""
+
+    def __init__(self, **kw):
+        opts = [['StrucId', CMQC.MQSRO_STRUC_ID, '4s'],
+                ['Version', CMQC.MQSRO_VERSION_1, MQLONG_TYPE],
+                ['Options', CMQC.MQSRO_FAIL_IF_QUIESCING, MQLONG_TYPE],
+                ['NumPubs', 0, MQLONG_TYPE]]
+
+        apply(MQOpts.__init__, (self, tuple(opts)), kw)
+
+
+
 #
 # A utility to convert a MQ constant to its string mnemonic by groping
 # a module dictonary
@@ -808,6 +1227,9 @@ class QueueManager:
         self.__handle = rv[0]
         self.__name = name
 
+    # Backward compatibility
+    connect_with_options = connectWithOptions
+
     def connectTCPClient(self, name, cd, channelName, connectString):
         """connectTCPClient(name, cd, channelName, connectString)
 
@@ -822,6 +1244,9 @@ class QueueManager:
         cd.TransportType = CMQC.MQXPT_TCP
         self.connectWithOptions(name, cd)
 
+    # Backward compatibility
+    connect_tcp_client = connectTCPClient
+
     def disconnect(self):
         """disconnect()
 
@@ -843,6 +1268,9 @@ class QueueManager:
         else:
             raise PYIFError('not connected')
 
+    # Backward compatibility
+    get_handle = getHandle
+
     def begin(self):
         """begin()
 
@@ -925,6 +1353,23 @@ class QueueManager:
             raise MQMIError(rv[-2], rv[-1])
         return rv[0]
 
+    def _is_connected(self):
+        """ Try pinging the queue manager in order to see whether the application
+        is connected to it. Note that the method is merely a convienece wrapper
+        around MQCMD_PING_Q_MGR, in particular, there's still possibility that
+        the app will disconnect between checking QueueManager.is_connected
+        and the next MQ call.
+        """
+        pcf = PCFExecute(self)
+        try:
+            pcf.MQCMD_PING_Q_MGR()
+        except Exception, e:
+            return False
+        else:
+            return True
+
+    is_connected = property(_is_connected)
+
 
 # Some support functions for Queue ops.
 def makeQDesc(qDescOrString):
@@ -934,6 +1379,7 @@ def makeQDesc(qDescOrString):
     else:
         return qDescOrString
 
+make_q_desc = makeQDesc
 
 def commonQArgs(*opts):
     "Process args common to put/get/put1. Module Private."
@@ -950,6 +1396,8 @@ def commonQArgs(*opts):
         mDesc = md()
     return(mDesc, pgOpts)
 
+common_q_args = commonQArgs
+
 
 class Queue:
 
@@ -1072,6 +1520,35 @@ class Queue:
         mDesc.unpack(rv[0])
         putOpts.unpack(rv[1])
 
+    def put_rfh2(self, msg, *opts):
+        """put_rfh2(msg[, mDesc ,putOpts, [rfh2_header, ]])
+
+        Put a RFH2 message. opts[2] is a list of RFH2 headers.
+        MQMD and RFH2's must be correct.
+
+        """
+
+        rfh2_buff = ""
+        if len(opts) >= 3:
+            if opts[2] is not None:
+                if not isinstance(opts[2], list):
+                    raise exceptions.TypeError("Third item of opts should " \
+                                               "be a list.")
+                encoding = CMQC.MQENC_NATIVE
+                if opts[0] is not None:
+                    mqmd = opts[0]
+                    encoding = mqmd["Encoding"]
+
+                for rfh2_header in opts[2]:
+                    if rfh2_header is not None:
+                        rfh2_buff = rfh2_buff + rfh2_header.pack(encoding)
+                        encoding = rfh2_header["Encoding"]
+
+                msg = rfh2_buff + msg
+            self.put(msg, *opts[0:2])
+        else:
+            self.put(msg, *opts)
+
     def get(self, maxLength = None, *opts):
         """get([maxLength [, mDesc, getOpts]])
 
@@ -1141,8 +1618,42 @@ class Queue:
             raise MQMIError(rv[-2], rv[-1])
         mDesc.unpack(rv[1])
         getOpts.unpack(rv[2])
+
         return rv[0]
 
+    def get_rfh2(self, max_length=None, *opts):
+        """get_rfh2([maxLength [, mDesc, getOpts, [rfh2_header_1, ]]])
+
+        Get a message and attempt to unpack the rfh2 headers.
+        opts[2] should be a empty list.
+        Unpacking only attempted if Format in previous header is
+        CMQC.MQFMT_RF_HEADER_2.
+
+        """
+
+        if len(opts) >= 3:
+            if opts[2] is not None:
+                if not isinstance(opts[2], list):
+                    raise exceptions.TypeError("Third item of opts should " \
+                                               "be a list.")
+
+                msg = self.get(max_length, *opts[0:2])
+                mqmd = opts[0]
+                rfh2_headers = []
+                #If format is not CMQC.MQFMT_RF_HEADER_2 then do not parse.
+                format = mqmd["Format"]
+                while format == CMQC.MQFMT_RF_HEADER_2:
+                    rfh2_header = RFH2()
+                    rfh2_header.unpack(msg)
+                    rfh2_headers.append(rfh2_header)
+                    msg = msg[rfh2_header["StrucLength"]:]
+                    format = rfh2_header["Format"]
+                opts[2].extend(rfh2_headers)
+        else:
+            msg = self.get(max_length, *opts)
+
+        return msg
+
     def close(self, options = CMQC.MQCO_NONE):
         """close([options])
 
@@ -1181,6 +1692,362 @@ class Queue:
         if rv[1]:
             raise MQMIError(rv[-2], rv[-1])
 
+    def set_handle(self, queue_handle):
+        """set_handle(queue_handle)
+
+        Sets the queue handle in the case when a handl;e was returned from a
+        previous MQ call.
+
+        """
+
+        self.__qHandle = queue_handle
+
+    def get_handle(self):
+        """get_handle()
+
+        Get the queue handle.
+
+        """
+
+        return self.__qHandle
+
+#Publish Subscribe support - Hannes Wagener 201
+class Topic:
+    """Topic(queue_manager, topic_name, topic_string, topic_desc, open_opts)
+
+    Topic encapsulates all the Topic I/O operations, including
+    publish/subscribe.  A QueueManager object must be already
+    connected. The Topic may be opened implicitly on construction, or
+    the open may be deferred until a call to open(), pub() or
+    (The same as for Queue).
+
+    The Topic to open is identified either by a topic name and/or a topic
+    string (in which case a default MQOD structure is created using
+    those names), or by passing a ready constructed MQOD class.
+    Refer to the "Using topic strings" section in the MQ7 Information Center
+    for an explanation of how the topic name and topic string is combined
+    to identify a particular topic.
+
+    """
+
+    def __real_open(self):
+        """real_open()
+
+        Really open the topic.  Only do this in pub()?
+
+        """
+
+        if self.__topic_desc == None:
+            raise PYIFError, 'The Topic Descriptor has not been set.'
+
+        rv = pymqe.MQOPEN(self.__queue_manager.getHandle(),
+                          self.__topic_desc.pack(), self.__open_opts)
+        if rv[-2]:
+            raise MQMIError(rv[-2], rv[-1])
+
+        self.__topic_handle = rv[0]
+        self.__topic_desc.unpack(rv[1])
+
+    def __init__(self, queue_manager, topic_name=None, topic_string=None, topic_desc=None, open_opts=None):
+        """Topic(queue_manager, [topic_name, topic_string, topic_desc [,open_opts]])
+
+        Associate a Topic instance with the QueueManager object 'queue_manager'
+        and optionally open the Topic.
+
+        If topic_desc is passed ignore topic_string and topic_name.
+
+        If open_opts is passed, it specifies topic open options, and
+        the topic is opened immediately. If open_opts is not passed,
+        the queue open is deferred to a subsequent call to open(),
+        pub().
+
+        The following table clarifies when the Topic is opened:
+
+        topic_desc  open_opts   When opened
+             N       N       open()
+             Y       N       open() or pub()
+             Y       Y       Immediately
+        """
+
+        self.__queue_manager = queue_manager
+        self.__topic_handle = None
+        self.__topic_desc = topic_desc
+        self.__open_opts = open_opts
+
+        self.topic_name = topic_name
+        self.topic_string = topic_string
+
+        if self.__topic_desc:
+            if self.__topic_desc["ObjectType"] is not CMQC.MQOT_TOPIC:
+                raise PYIFError, 'The Topic Descriptor ObjectType is not '\
+                                 'MQOT_TOPIC.'
+            if self.__topic_desc["Version"] is not CMQC.MQOD_VERSION_4:
+                raise PYIFError, 'The Topic Descriptor Version is not '\
+                                 'MQOD_VERSION_4.'
+        else:
+            self.__topic_desc = self.__create_topic_desc(topic_name, topic_string)
+
+        if self.__open_opts:
+            self.__real_open()
+
+    def __create_topic_desc(self, topic_name, topic_string):
+        """__create_topic(topic_name, topic_string)
+
+        Creates a topic object descriptor from a given topic_name/topic_string.
+
+        """
+
+        topic_desc = od()
+        topic_desc["ObjectType"] = CMQC.MQOT_TOPIC
+        topic_desc["Version"] = CMQC.MQOD_VERSION_4
+
+        if topic_name:
+            topic_desc["ObjectName"] = topic_name
+
+        if topic_string:
+            topic_desc.set_vs("ObjectString", topic_string, 0, 0, 0)
+
+        return topic_desc
+
+    def __del__(self):
+        """__del__()
+
+        Close the Topic, if it has been opened."""
+        try:
+            if self.__topic_handle:
+                self.close()
+        except:
+            pass
+
+    def open(self, topic_name=None, topic_string=None, topic_desc=None, open_opts=None):
+        """open(topic_name, topic_string, topic_desc, open_opts)
+
+        Open the Topic specified by topic_desc or create a object descriptor
+        from topic_name and topic_string.
+        If open_opts is passed, it defines the
+        Topic open options, and the Topic is opened immediately. If
+        open_opts is not passed, the Topic open is deferred until a
+        subsequent pub() call.
+
+        """
+
+        if self.__topic_handle:
+            raise PYIFError('The Topic is already open.')
+
+        if topic_name:
+            self.topic_name = topic_name
+
+        if topic_string:
+            self.topic_string = topic_string
+
+        if topic_desc:
+            self.__topic_desc = topic_desc
+        else:
+            self.__topic_desc = self.__create_topic_desc(self.topic_name, self.topic_string)
+
+        if open_opts:
+            self.__open_opts = open_opts
+            self.__real_open()
+
+    def pub(self, msg, *opts):
+        """pub(msg[, msg_desc ,put_opts])
+
+        Publish the string buffer 'msg' to the Topic. If the Topic is not
+        already open, it is opened now. with the option 'MQOO_OUTPUT'.
+
+        msg_desc is the pymqi.md() MQMD Message Descriptor for the
+        message. If it is not passed, or is None, then a default md()
+        object is used.
+
+        put_opts is the pymqi.pmo() MQPMO Put Message Options structure
+        for the put call. If it is not passed, or is None, then a
+        default pmo() object is used.
+
+        If msg_desc and/or put_opts arguments were supplied, they may be
+        updated by the put operation.
+
+        """
+
+        msg_desc, put_opts = apply(commonQArgs, opts)
+
+        if put_opts == None:
+            put_opts = pmo()
+
+        # If queue open was deferred, open it for put now
+        if not self.__topic_handle:
+            self.__open_opts = CMQC.MQOO_OUTPUT
+            self.__real_open()
+        # Now send the message
+        rv = pymqe.MQPUT(self.__queue_manager.getHandle(), self.__topic_handle, msg_desc.pack(),
+                         put_opts.pack(), msg)
+        if rv[-2]:
+            raise MQMIError(rv[-2], rv[-1])
+
+        msg_desc.unpack(rv[0])
+        put_opts.unpack(rv[1])
+
+    def sub(self, *opts):
+        """sub(sub_desc, sub_queue)
+
+        Subscribe to the topic and return a Subscription object.
+        A subscription to a topic can be made using an existing queue, either
+        by pasing a Queue object or a string at which case the queue will
+        be opened with default options.
+
+        """
+
+        sub_desc = None
+        if len(opts) > 0:
+            sub_desc = opts[0]
+
+        sub_queue = None
+        if len(opts) > 1:
+            sub_queue = opts[1]
+
+        sub = Subscription(self.__queue_manager)
+        sub.sub(sub_desc=sub_desc, sub_queue=sub_queue, topic_name=self.topic_name, topic_string=self.topic_string)
+
+        return sub
+
+    def close(self, options = CMQC.MQCO_NONE):
+        """close([options])
+
+        Close the topic, using options.
+
+        """
+
+        if not self.__topic_handle:
+            raise PYIFError('Topic not open.')
+
+        rv = pymqe.MQCLOSE(self.__queue_manager.getHandle(), self.__topic_handle, options)
+        if rv[0]:
+            raise MQMIError(rv[-2], rv[-1])
+
+        self.__topic_handle = None
+        self.__topic_desc = None
+        self.__open_opts = None
+
+
+class Subscription:
+    """Subscription(queue_manager, sub_descm sub_name, sub_queue, topic_name,
+                    topic_string)
+
+    Subscription encapsulates a subscription to a topic.
+
+    """
+    def __init__(self, queue_manager, sub_desc=None, sub_name=None,
+                 sub_queue=None, sub_opts=None, topic_name=None, topic_string=None):
+        self.__queue_manager = queue_manager
+        self.sub_queue = sub_queue
+        self.__sub_desc = sub_desc
+        self.sub_name = sub_name
+        self.sub_opts = sub_opts
+        self.topic_name = topic_name
+        self.topic_string = topic_string
+
+        if self.__sub_desc:
+            self.sub(sub_desc=self.__sub_desc)
+
+    def get_sub_queue(self):
+        """get_sub_queue()
+
+        Return the subscription queue.
+
+        """
+        return self.sub_queue
+
+
+    def get(self, max_length=None, *opts):
+        """
+
+        Get a publication from the Queue.
+
+        """
+
+        return self.sub_queue.get(max_length, *opts)
+
+    def sub(self, sub_desc=None, sub_queue=None, sub_name=None, sub_opts=None,
+            topic_name=None, topic_string=None):
+        """sub(sub_desc, sub_queue)
+
+        Subscribe to a topic, alter or resume a subscription.
+        Executes the MQSUB call with parameters.
+        The subscription queue can be either passed as a Queue object or a
+        Queue object handle.
+
+        """
+
+        if topic_name:
+            self.topic_name = topic_name
+        if topic_string:
+            self.topic_string = topic_string
+        if sub_name:
+            self.sub_name = sub_name
+
+        if sub_desc:
+            if not isinstance(sub_desc, SD):
+                raise exceptions.TypeError("sub_desc must be a " \
+                                               "SD(sub descriptor) object.")
+        else:
+            sub_desc = SD()
+            if sub_opts:
+                sub_desc["Options"] = sub_opts
+            else:
+                sub_desc["Options"] = CMQC.MQSO_CREATE + \
+                                      CMQC.MQSO_NON_DURABLE + \
+                                      CMQC.MQSO_MANAGED
+            if self.sub_name:
+                sub_desc.set_vs("SubName", self.sub_name)
+            if self.topic_name:
+                sub_desc["ObjectName"] = self.topic_name
+            if self.topic_string:
+                sub_desc.set_vs("ObjectString", self.topic_string)
+        self.__sub_desc = sub_desc
+
+        sub_queue_handle = CMQC.MQHO_NONE
+        if sub_queue:
+            if isinstance(sub_queue, Queue):
+                sub_queue_handle = sub_queue.get_handle()
+            else:
+                sub_queue_handle = sub_queue
+
+        rv = pymqe.MQSUB(self.__queue_manager.getHandle(), sub_desc.pack(),
+                         sub_queue_handle)
+
+        if rv[-2]:
+            raise MQMIError(rv[-2], rv[-1])
+
+        sub_desc.unpack(rv[0])
+        self.__sub_desc = sub_desc
+        self.sub_queue = Queue(self.__queue_manager)
+        self.sub_queue.set_handle(rv[1])
+        self.__sub_handle = rv[2]
+
+    def close(self, sub_close_options=CMQC.MQCO_NONE, close_sub_queue=False, close_sub_queue_options=CMQC.MQCO_NONE):
+
+        if not self.__sub_handle:
+            raise PYIFError('Subscription not open.')
+
+        rv = pymqe.MQCLOSE(self.__queue_manager.getHandle(), self.__sub_handle, sub_close_options)
+        if rv[0]:
+            raise MQMIError(rv[-2], rv[-1])
+
+        self.__sub_handle = None
+        self.__sub_desc = None
+        self.__open_opts = None
+
+        if close_sub_queue:
+            self.sub_queue.close(close_sub_queue_options)
+
+    def __del__(self):
+        """__del__()
+
+        Close the Subscription, if it has been opened."""
+        try:
+            if self.__sub_handle:
+                self.close()
+        except:
+            pass
 
 #
 # This piece of magic shamelessly plagiarised from xmlrpclib.py. It
@@ -1278,6 +2145,34 @@ class PCFExecute(QueueManager):
                 rv[k] = rawDict[k]
         return rv
 
+    # Backward compatibility
+    stringify_keys = stringifyKeys
+
+class ByteString(object):
+    """ A simple wrapper around string values, suitable for passing into PyMQI
+    calls wherever IBM's docs state a 'byte string' object should be passed in.
+    """
+    def __init__(self, value):
+        self.value = value
+        self.pymqi_byte_string = True
+
+    def __len__(self):
+        return len(self.value)
+    
+def connect(queue_manager, channel=None, conn_info=None):
+    """ A convenience wrapper for connecting to MQ queue managers. If given the
+    'queue_manager' parameter only, will try connecting to it in bindings mode.
+    If given both 'channel' and 'conn_info' will connect in client mode.
+    A pymqi.QueueManager is returned on successfully establishing a connection.
+    """
+    if queue_manager and channel and conn_info:
+        qmgr = QueueManager(None)
+        qmgr.connect_tcp_client(queue_manager, CD(), channel, conn_info)
+        return qmgr
+    
+    elif queue_manager:
+        qmgr = pymqi.QueueManager(queue_manager)
+        return qmgr
 
 #######################################################################
 #
